from dataclasses import dataclass
from datetime import timedelta
from enum import StrEnum


class ExecutorClass(StrEnum):
    spin_up_4min__gpu_24gb = "spin_up-4min.gpu-24gb"
    always_on__gpu_24gb = "always_on.gpu-24gb"
    # always_on__cpu_16c__ram_64gb = "always_on.cpu-16c.ram-64gb"
    # always_on__gpu_80gb = "always_on.gpu-80gb"
    # always_on__gpu_24gb__docker_cached_facilitator = "always_on.gpu-24gb.docker_cached-facilitator"


@dataclass
class ExecutorClassSpec:
    # mostly for user consumption, intended usage, ...
    description: str | None = None
    has_gpu: bool | None = None
    cpu_cores: int | None = None
    ram_gb: int | None = None
    storage_gb: int | None = None
    gpu_vram_gb: int | None = None
    spin_up_time: int | None = None  # seconds
    docker_cached_images: list[str] | None = None
    # requirements which can't be easily standardized
    additional_requirements: str | None = None


EXECUTOR_CLASS = {
    ExecutorClass.spin_up_4min__gpu_24gb: ExecutorClassSpec(
        description="cost effective (started on demand), generic GPU machine",
        has_gpu=True,
        gpu_vram_gb=24,
        spin_up_time=int(timedelta(minutes=4).total_seconds()),
    ),
    ExecutorClass.always_on__gpu_24gb: ExecutorClassSpec(
        description="always on, generic GPU machine",
        has_gpu=True,
        gpu_vram_gb=24,
        spin_up_time=0,
    ),
    # ExecutorClass.always_on__cpu_16c__ram_64gb: ExecutorClassSpec(
    #     cpu_cores=16,
    #     ram_gb=64,
    #     spin_up_time=0,
    # ),
    # ExecutorClass.always_on__gpu_80gb: ExecutorClassSpec(
    #     has_gpu=True,
    #     gpu_vram_gb=80,
    #     spin_up_time=0,
    # ),
    # ExecutorClass.always_on__gpu_24gb__docker_cached_facilitator: ExecutorClassSpec(
    #     has_gpu=True,
    #     gpu_vram_gb=24,
    #     spin_up_time=0,
    #     docker_cached_images=[
    #         "backenddevelopersltd/compute-horde-job-huggingface",
    #         "backenddevelopersltd/compute-horde-job-ffmpeg:vidstab-v0",
    #     ],
    # ),
}


# this leaves around 1 min for synthetic job to complete
MAX_EXECUTOR_TIMEOUT = timedelta(minutes=4).total_seconds()

DEFAULT_EXECUTOR_CLASS = ExecutorClass.spin_up_4min__gpu_24gb
DEFAULT_EXECUTOR_TIMEOUT = EXECUTOR_CLASS[DEFAULT_EXECUTOR_CLASS].spin_up_time



import csv
import io

import pytest

from compute_horde.receipts import ReceiptFetchError, get_miner_receipts


def receipts_helper(mocked_responses, receipts, miner_keypair):
    buf = io.StringIO()
    csv_writer = csv.writer(buf)
    csv_writer.writerow(
        [
            "job_uuid",
            "miner_hotkey",
            "validator_hotkey",
            "time_started",
            "time_took_us",
            "score_str",
            "validator_signature",
            "miner_signature",
        ]
    )
    for receipt in receipts:
        csv_writer.writerow(
            [
                receipt.payload.job_uuid,
                receipt.payload.miner_hotkey,
                receipt.payload.validator_hotkey,
                receipt.payload.time_started.isoformat(),
                receipt.payload.time_took_us,
                receipt.payload.score_str,
                receipt.validator_signature,
                receipt.miner_signature,
            ]
        )

    buf.seek(0)
    mocked_responses.get("http://127.0.0.1:8000/receipts/receipts.csv", body=buf.read())
    return get_miner_receipts(miner_keypair.ss58_address, "127.0.0.1", 8000)


def receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair):
    got_receipts = receipts_helper(mocked_responses, receipts, miner_keypair)
    # only the valid receipt should be stored
    assert len(got_receipts) == 1
    assert got_receipts[0] == receipts[0]


def test__get_miner_receipts__happy_path(mocked_responses, receipts, miner_keypair):
    got_receipts = receipts_helper(mocked_responses, receipts, miner_keypair)
    assert len(got_receipts) == 2
    for receipt in receipts:
        got_receipt = [x for x in got_receipts if x.payload.job_uuid == receipt.payload.job_uuid][0]
        assert got_receipt == receipt


def test__get_miner_receipts__invalid_receipt_skipped(mocked_responses, receipts, miner_keypair):
    """
    Populate all the fields of one csv row with "invalid" :D
    """

    class Mock:
        def __str__(self):
            return "invalid"

        isoformat = __str__

        def __getattr__(self, item):
            return Mock()

    receipts[1] = Mock()
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__miner_hotkey_mismatch_skipped(
    mocked_responses, receipts, miner_keypair, keypair
):
    receipts[1].payload.miner_hotkey = keypair.ss58_address
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__invalid_miner_signature_skipped(
    mocked_responses, receipts, miner_keypair
):
    receipts[1].miner_signature = f"0x{miner_keypair.sign('bla').hex()}"
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__invalid_validator_signature_skipped(
    mocked_responses, receipts, miner_keypair
):
    receipts[1].validator_signature = f"0x{miner_keypair.sign('bla').hex()}"
    receipts_one_skipped_helper(mocked_responses, receipts, miner_keypair)


def test__get_miner_receipts__no_receipts(mocked_responses, miner_keypair):
    mocked_responses.get("http://127.0.0.1:8000/receipts/receipts.csv", status=404)
    with pytest.raises(ReceiptFetchError):
        get_miner_receipts(miner_keypair.ss58_address, "127.0.0.1", 8000)

    with pytest.raises(ReceiptFetchError):
        get_miner_receipts(miner_keypair.ss58_address, "127.0.0.1", 8001)



import datetime

import pytest
import responses
from bittensor import Keypair

from compute_horde.mv_protocol.validator_requests import ReceiptPayload
from compute_horde.receipts import Receipt


@pytest.fixture
def keypair():
    return Keypair.create_from_mnemonic(
        "edit evoke caught tunnel harsh plug august group enact cable govern immense"
    )


@pytest.fixture
def validator_keypair():
    return Keypair.create_from_mnemonic(
        "slot excuse valid grief praise rifle spoil auction weasel glove pen share"
    )


@pytest.fixture
def miner_keypair():
    return Keypair.create_from_mnemonic(
        "almost fatigue race slim picnic mass better clog deal solve already champion"
    )


@pytest.fixture
def receipts(validator_keypair, miner_keypair):
    payload1 = ReceiptPayload(
        job_uuid="0d89161e-65e4-46ad-bed8-ecfec1cc3c6b",
        miner_hotkey=miner_keypair.ss58_address,
        validator_hotkey=validator_keypair.ss58_address,
        time_started=datetime.datetime(2024, 1, 1, tzinfo=datetime.UTC),
        time_took_us=1_000_000,
        score_str="1.00",
    )
    receipt1 = Receipt(
        payload=payload1,
        validator_signature=f"0x{validator_keypair.sign(payload1.blob_for_signing()).hex()}",
        miner_signature=f"0x{miner_keypair.sign(payload1.blob_for_signing()).hex()}",
    )

    payload2 = ReceiptPayload(
        job_uuid="3342460e-4a99-438b-8757-795f4cb348dd",
        miner_hotkey=miner_keypair.ss58_address,
        validator_hotkey=validator_keypair.ss58_address,
        time_started=datetime.datetime(2024, 1, 2, tzinfo=datetime.UTC),
        time_took_us=2_000_000,
        score_str="2.00",
    )
    receipt2 = Receipt(
        payload=payload2,
        validator_signature=f"0x{validator_keypair.sign(payload2.blob_for_signing()).hex()}",
        miner_signature=f"0x{miner_keypair.sign(payload2.blob_for_signing()).hex()}",
    )

    return [receipt1, receipt2]


@pytest.fixture
def mocked_responses():
    with responses.RequestsMock() as rsps:
        yield rsps



import enum
import re
from typing import Annotated, Literal
from urllib.parse import urlparse

import pydantic
from pydantic import Field

SAFE_DOMAIN_REGEX = re.compile(r".*")


class VolumeType(str, enum.Enum):
    inline = "inline"
    zip_url = "zip_url"
    single_file = "single_file"
    multi_volume = "multi_volume"

    def __str__(self):
        return str.__str__(self)


class InlineVolume(pydantic.BaseModel):
    volume_type: Literal[VolumeType.inline] = VolumeType.inline
    contents: str
    relative_path: str | None = None

    def is_safe(self) -> bool:
        return True


class ZipUrlVolume(pydantic.BaseModel):
    volume_type: Literal[VolumeType.zip_url] = VolumeType.zip_url
    contents: str  # backwards compatible
    relative_path: str | None = Field(default=None)

    def is_safe(self) -> bool:
        domain = urlparse(self.contents).netloc
        if SAFE_DOMAIN_REGEX.fullmatch(domain):
            return True
        return False


class SingleFileVolume(pydantic.BaseModel):
    volume_type: Literal[VolumeType.single_file] = VolumeType.single_file
    url: str
    relative_path: str

    def is_safe(self) -> bool:
        domain = urlparse(self.url).netloc
        if SAFE_DOMAIN_REGEX.fullmatch(domain):
            return True
        return False


class MultiVolume(pydantic.BaseModel):
    volume_type: Literal[VolumeType.multi_volume] = VolumeType.multi_volume
    volumes: list[
        Annotated[
            InlineVolume | ZipUrlVolume | SingleFileVolume, Field(discriminator="volume_type")
        ]
    ]

    def is_safe(self) -> bool:
        return all(volume.is_safe() for volume in self.volumes)


Volume = Annotated[
    InlineVolume | ZipUrlVolume | SingleFileVolume | MultiVolume,
    Field(discriminator="volume_type"),
]



import enum
from typing import Self

from pydantic import model_validator

from ..base.output_upload import OutputUpload, OutputUploadType  # noqa
from ..base.volume import Volume, VolumeType
from ..base_requests import BaseRequest, JobMixin


class RequestType(enum.Enum):
    V0PrepareJobRequest = "V0PrepareJobRequest"
    V0RunJobRequest = "V0RunJobRequest"
    GenericError = "GenericError"


class BaseMinerRequest(BaseRequest):
    message_type: RequestType


class V0InitialJobRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0PrepareJobRequest
    base_docker_image_name: str | None = None
    timeout_seconds: int | None = None
    volume_type: VolumeType | None = None


class V0JobRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0RunJobRequest
    docker_image_name: str | None = None
    raw_script: str | None = None
    docker_run_options_preset: str
    docker_run_cmd: list[str]
    volume: Volume | None = None
    output_upload: OutputUpload | None = None

    @model_validator(mode="after")
    def validate_at_least_docker_image_or_raw_script(self) -> Self:
        if not (bool(self.docker_image_name) or bool(self.raw_script)):
            raise ValueError("Expected at least one of `docker_image_name` or `raw_script`")
        return self


class GenericError(BaseMinerRequest):
    message_type: RequestType = RequestType.GenericError
    details: str | None = None



import logging

import pytest

from ..base.volume import ZipUrlVolume

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "url, expected",
    [
        ("http://199.199.99.99:3000/file.zip", True),
        ("http://google.com/file.zip", True),
        ("https://drive.google.com/uc?export=download&id=1qqq", True),
        ("https://github.com/qqquser/hordetest/blob/main/image.zip", True),
        ("https://horde-model-job.s3.amazonaws.com/test.zip", True),
        (
            "https://raw.githubusercontent.com/backend-developers-ltd/ComputeHorde/master/file.zip",
            True,
        ),
        (
            "https://spanish-translator.s3.amazonaws.com/file.zip?X-Amz-Algorithm=SHA256&X-Amz-Credential=qqq",
            True,
        ),
        ("https://fake.s3.amazonaws.com.fake.com/test.zip", True),
        ("https://something.r2.cloudflarestorage.com/test.zip", True),
    ],
)
def test_volume_is_safe(url: str, expected: bool):
    volume = ZipUrlVolume(contents=url)
    assert volume.is_safe() == expected



import abc
import logging
import os
import select
import subprocess
import time
import typing
from threading import Thread

import pytest

logger = logging.getLogger(__name__)


class ActiveSubnetworkBaseTest(abc.ABC):
    miner_process = None
    validator_process = None
    validator_stdout_thread: Thread | None = None
    validator_stderr_thread: Thread | None = None
    miner_stdout_thread: Thread | None = None
    miner_stderr_thread: Thread | None = None

    @classmethod
    def miner_preparation_tasks(cls):
        """
        Arbitrary preparation tasks to be performed before starting the miner.
        """

    @classmethod
    @abc.abstractmethod
    def validator_path_and_args(cls) -> list[str]:
        """
        Return the path to validator start script and all arguments required to start it, e.g.
        `validator.py --opt1=8 --opt2=9`
        """
        ...

    @classmethod
    @abc.abstractmethod
    def miner_path_and_args(cls) -> list[str]:
        """
        Return the path to miner start script and all arguments required to start it, e.g.
        `miner.py --opt1=8 --opt2=9`
        """
        ...

    @classmethod
    @abc.abstractmethod
    def check_if_validator_is_up(cls) -> bool:
        """
        Check if the validator process is up and ready to accept requests. This can be achieved by making a network,
        reading logs etc.
        """
        ...

    @classmethod
    @abc.abstractmethod
    def check_if_miner_is_up(cls) -> bool:
        """
        Check if the miner process is up and ready to accept requests. This can be achieved by making a network, reading
        logs etc.
        """
        ...

    @classmethod
    @abc.abstractmethod
    def miner_environ(cls) -> dict[str, str]:
        """
        Additional env vars for starting the miner.
        """
        ...

    @classmethod
    @abc.abstractmethod
    def validator_environ(cls) -> dict[str, str]:
        """
        Additional env vars for starting the validator.
        """
        ...

    @classmethod
    def make_log_reader(cls, io: typing.IO, name: str, process: subprocess.Popen):
        def read_logs():
            while True:
                r, _, _ = select.select([io], [], [], 1)
                if process.poll() is not None:
                    return
                if r:
                    line = io.readline()
                    if not line:
                        return
                    logger.info(f"{name}: {line.decode()[:-1]}")

        return read_logs

    @classmethod
    def start_process(cls, args, additional_env: dict[str, str]):
        return subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            start_new_session=True,
            env={**os.environ, **additional_env},
        )

    @classmethod
    def wait_for_process_start(cls, process_name, probe_function, process: subprocess.Popen):
        for i in range(300):
            if probe_function():
                return
            if process.poll() is not None:
                break
            time.sleep(0.1)
        raise RuntimeError(
            f"Process {process_name} did not start\n"
            f"stdout={process.stdout.read()}\n\n"
            f"stderr={process.stderr.read()}"
        )

    @classmethod
    @pytest.fixture(autouse=True, scope="session")
    def start_validator_and_miner(cls):
        logger.info("Starting miner")
        cls.miner_preparation_tasks()
        cls.miner_process = cls.start_process(cls.miner_path_and_args(), cls.miner_environ())
        cls.miner_stdout_thread = Thread(
            target=cls.make_log_reader(cls.miner_process.stdout, "miner stdout", cls.miner_process)
        )
        cls.miner_stderr_thread = Thread(
            target=cls.make_log_reader(cls.miner_process.stderr, "miner stderr", cls.miner_process)
        )
        cls.miner_stdout_thread.start()
        cls.miner_stderr_thread.start()

        logger.info("Starting validator")
        cls.validator_process = cls.start_process(
            cls.validator_path_and_args(), cls.validator_environ()
        )
        cls.validator_stdout_thread = Thread(
            target=cls.make_log_reader(
                cls.validator_process.stdout, "validator stdout", cls.validator_process
            )
        )
        cls.validator_stderr_thread = Thread(
            target=cls.make_log_reader(
                cls.validator_process.stderr, "validator stderr", cls.validator_process
            )
        )
        cls.validator_stdout_thread.start()
        cls.validator_stderr_thread.start()

        logger.info("Waiting for validator to start")
        cls.wait_for_process_start("validator", cls.check_if_validator_is_up, cls.validator_process)
        logger.info("Waiting for miner to start")
        cls.wait_for_process_start("miner", cls.check_if_miner_is_up, cls.miner_process)
        yield
        if cls.miner_process:
            logger.info("Killing miner")
            os.killpg(os.getpgid(cls.miner_process.pid), 9)
        if cls.validator_process:
            logger.info("Killing validator")
            os.killpg(os.getpgid(cls.validator_process.pid), 9)



import abc
import enum
import json

import pydantic


class ValidationError(Exception):
    def __init__(self, msg):
        self.msg = msg

    @classmethod
    def from_json_decode_error(cls, exc: json.JSONDecodeError):
        return cls(exc.args[0])

    @classmethod
    def from_pydantic_validation_error(cls, exc: pydantic.ValidationError):
        return cls(json.dumps(exc.json()))

    def __repr__(self):
        return f"{type(self).__name__}({self.msg})"


def all_subclasses(cls: type):
    for subcls in cls.__subclasses__():
        yield subcls
        yield from all_subclasses(subcls)


base_class_to_request_type_mapping = {}


class BaseRequest(pydantic.BaseModel, abc.ABC):
    message_type: enum.Enum

    @classmethod
    def type_to_model(cls, type_: enum.Enum) -> type["BaseRequest"]:
        mapping = base_class_to_request_type_mapping.get(cls)
        if not mapping:
            mapping = {}
            for klass in all_subclasses(cls):
                if not (message_type := klass.model_fields.get("message_type")):
                    continue
                if not message_type.default:
                    continue
                mapping[message_type.default] = klass
            base_class_to_request_type_mapping[cls] = mapping

        return mapping[type_]

    @classmethod
    def parse(cls, str_: str):
        try:
            json_ = json.loads(str_)
        except json.JSONDecodeError as exc:
            raise ValidationError.from_json_decode_error(exc)

        try:
            base_model_object = cls.model_validate(json_)
        except pydantic.ValidationError as exc:
            raise ValidationError.from_pydantic_validation_error(exc)

        target_model = cls.type_to_model(base_model_object.message_type)

        try:
            return target_model.model_validate(json_)
        except pydantic.ValidationError as exc:
            raise ValidationError.from_pydantic_validation_error(exc)


class JobMixin(pydantic.BaseModel):
    job_uuid: str



import enum
import re
from collections.abc import Mapping
from typing import Annotated, Literal
from urllib.parse import urlparse

import pydantic
from pydantic import Field

SAFE_DOMAIN_REGEX = re.compile(r".*")


class OutputUploadType(str, enum.Enum):
    zip_and_http_post = "zip_and_http_post"
    zip_and_http_put = "zip_and_http_put"
    multi_upload = "multi_upload"
    single_file_post = "single_file_post"
    single_file_put = "single_file_put"

    def __str__(self):
        return str.__str__(self)


class ZipAndHttpPostUpload(pydantic.BaseModel):
    output_upload_type: Literal[OutputUploadType.zip_and_http_post] = (
        OutputUploadType.zip_and_http_post
    )
    url: str
    form_fields: Mapping[str, str] | None = None


class ZipAndHttpPutUpload(pydantic.BaseModel):
    output_upload_type: Literal[OutputUploadType.zip_and_http_put] = (
        OutputUploadType.zip_and_http_put
    )
    url: str
    # TODO: PUT implementation does not support it - find out why and clean up
    # form_fields: Mapping[str, str] | None = None


class SingleFilePostUpload(pydantic.BaseModel):
    output_upload_type: Literal[OutputUploadType.single_file_post] = (
        OutputUploadType.single_file_post
    )
    url: str
    form_fields: Mapping[str, str] | None = None
    relative_path: str
    signed_headers: Mapping[str, str] | None = None

    def is_safe(self) -> bool:
        domain = urlparse(self.url).netloc
        if SAFE_DOMAIN_REGEX.fullmatch(domain):
            return True
        return False


class SingleFilePutUpload(pydantic.BaseModel):
    output_upload_type: Literal[OutputUploadType.single_file_put] = OutputUploadType.single_file_put
    url: str
    # TODO: PUT implementation does not support it - find out why and clean up
    # form_fields: Mapping[str, str] | None = None
    relative_path: str
    signed_headers: Mapping[str, str] | None = None

    def is_safe(self) -> bool:
        domain = urlparse(self.url).netloc
        if SAFE_DOMAIN_REGEX.fullmatch(domain):
            return True
        return False


SingleFileUpload = Annotated[
    SingleFilePostUpload | SingleFilePutUpload,
    Field(discriminator="output_upload_type"),
]


class MultiUpload(pydantic.BaseModel):
    output_upload_type: Literal[OutputUploadType.multi_upload] = OutputUploadType.multi_upload
    uploads: list[SingleFileUpload]
    # allow custom uploads for stdout and stderr
    system_output: ZipAndHttpPostUpload | ZipAndHttpPutUpload | None = None


OutputUpload = Annotated[
    ZipAndHttpPostUpload | ZipAndHttpPutUpload | MultiUpload,
    Field(discriminator="output_upload_type"),
]



import abc
import asyncio
import logging
import random
import time

import websockets

from compute_horde.base_requests import BaseRequest, ValidationError

logger = logging.getLogger(__name__)


class MinerConnectionError(Exception):
    pass


class AbstractMinerClient(abc.ABC):
    def __init__(self, loop: asyncio.AbstractEventLoop, miner_name: str):
        self.debounce_counter = 0
        self.max_debounce_count: int | None = 5  # set to None for unlimited debounce
        self.loop = loop
        self.miner_name = miner_name
        self.ws: websockets.WebSocketClientProtocol | None = None
        self.read_messages_task: asyncio.Task | None = None
        self.deferred_send_tasks: list[asyncio.Task] = []

    @abc.abstractmethod
    def miner_url(self) -> str: ...

    @abc.abstractmethod
    def accepted_request_type(self) -> type[BaseRequest]:
        pass

    @abc.abstractmethod
    def incoming_generic_error_class(self) -> type[BaseRequest]:
        pass

    @abc.abstractmethod
    def outgoing_generic_error_class(self) -> type[BaseRequest]:
        pass

    @abc.abstractmethod
    async def handle_message(self, msg: BaseRequest):
        """
        Handle the message based on its type or raise UnsupportedMessageReceived
        """
        ...

    async def __aenter__(self):
        await self.await_connect()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for t in self.deferred_send_tasks:
            t.cancel()

        if self.read_messages_task is not None and not self.read_messages_task.done():
            self.read_messages_task.cancel()

        if self.ws is not None and not self.ws.closed:
            await self.ws.close()

    async def _connect(self):
        return await websockets.connect(self.miner_url(), max_size=50 * (2**20))  # 50MB

    async def await_connect(self):
        start_time = time.time()
        while True:
            try:
                if (
                    self.max_debounce_count is not None
                    and self.debounce_counter > self.max_debounce_count
                ):
                    time_took = time.time() - start_time
                    raise MinerConnectionError(
                        f"Could not connect to miner {self.miner_name} after {self.max_debounce_count} tries"
                        f" in {time_took:0.2f} seconds"
                    )
                if self.debounce_counter:
                    sleep_time = self.sleep_time()
                    logger.info(
                        f"Retrying connection to miner {self.miner_name} in {sleep_time:0.2f}"
                    )
                    await asyncio.sleep(sleep_time)
                self.ws = await self._connect()
                self.read_messages_task = self.loop.create_task(self.read_messages())
                if self.debounce_counter:
                    logger.info(
                        f"Connected to miner {self.miner_name} after {self.debounce_counter + 1} attempts"
                    )
                return
            except (websockets.WebSocketException, OSError) as ex:
                self.debounce_counter += 1
                logger.info(f"Could not connect to miner {self.miner_name}: {str(ex)}")

    def sleep_time(self):
        return (2**self.debounce_counter) + random.random()

    async def ensure_connected(self):
        if self.ws is None or self.ws.closed:
            if self.read_messages_task is not None and not self.read_messages_task.done():
                self.read_messages_task.cancel()
            await self.await_connect()

    async def send_model(self, model: BaseRequest):
        while True:
            await self.ensure_connected()
            try:
                await self.ws.send(model.model_dump_json())
            except websockets.WebSocketException as ex:
                logger.error(f"Could not send to miner {self.miner_name}: {str(ex)}")
                await asyncio.sleep(1 + random.random())
                continue
            return

    def deferred_send_model(self, model: BaseRequest):
        task = self.loop.create_task(self.send_model(model))
        self.deferred_send_tasks.append(task)

    async def read_messages(self):
        while True:
            try:
                msg = await self.ws.recv()
            except websockets.WebSocketException as ex:
                self.debounce_counter += 1
                logger.info(f"Connection to miner {self.miner_name} lost: {str(ex)}")
                self.loop.create_task(self.await_connect())
                return

            try:
                msg = self.accepted_request_type().parse(msg)
            except ValidationError as ex:
                error_msg = f"Malformed message from miner {self.miner_name}: {str(ex)}"
                logger.info(error_msg)
                self.deferred_send_model(self.outgoing_generic_error_class()(details=error_msg))
                continue

            if isinstance(msg, self.incoming_generic_error_class()):
                try:
                    raise RuntimeError(
                        f"Received error message from miner {self.miner_name}: {msg.model_dump_json()}"
                    )
                except Exception:
                    logger.exception("")
                continue
            try:
                await self.handle_message(msg)
            except UnsupportedMessageReceived:
                error_msg = f"Unsupported message from miner {self.miner_name}"
                logger.exception(error_msg)
                self.deferred_send_model(self.outgoing_generic_error_class()(details=error_msg))
            else:
                if self.debounce_counter:
                    logger.info(
                        f"Receviced valid message from miner {self.miner_name} after {self.debounce_counter + 1} connection attempts"
                    )
                    self.debounce_counter = 0


class UnsupportedMessageReceived(Exception):
    def __init__(self, msg: BaseRequest):
        self.msg = msg

    def __str__(self):
        return f"{type(self).__name__}: {self.msg.model_dump_json()}"

    __repr__ = __str__



import datetime
import enum
import json
import re
from typing import Self

import pydantic
from pydantic import model_validator

from ..base.output_upload import OutputUpload, OutputUploadType  # noqa
from ..base.volume import Volume, VolumeType
from ..base_requests import BaseRequest, JobMixin
from ..executor_class import ExecutorClass
from ..utils import MachineSpecs, _json_dumps_default

SAFE_DOMAIN_REGEX = re.compile(r".*")


class RequestType(enum.Enum):
    V0AuthenticateRequest = "V0AuthenticateRequest"
    V0InitialJobRequest = "V0InitialJobRequest"
    V0MachineSpecsRequest = "V0MachineSpecsRequest"
    V0JobRequest = "V0JobRequest"
    V0ReceiptRequest = "V0ReceiptRequest"
    GenericError = "GenericError"


class BaseValidatorRequest(BaseRequest):
    message_type: RequestType


class AuthenticationPayload(pydantic.BaseModel):
    validator_hotkey: str
    miner_hotkey: str
    timestamp: int

    def blob_for_signing(self):
        # pydantic v2 does not support sort_keys anymore.
        return json.dumps(self.model_dump(), sort_keys=True)


class V0AuthenticateRequest(BaseValidatorRequest):
    message_type: RequestType = RequestType.V0AuthenticateRequest
    payload: AuthenticationPayload
    signature: str

    def blob_for_signing(self):
        return self.payload.blob_for_signing()


class V0InitialJobRequest(BaseValidatorRequest, JobMixin):
    message_type: RequestType = RequestType.V0InitialJobRequest
    executor_class: ExecutorClass | None = None
    base_docker_image_name: str | None = None
    timeout_seconds: int | None = None
    volume_type: VolumeType | None = None


class V0JobRequest(BaseValidatorRequest, JobMixin):
    message_type: RequestType = RequestType.V0JobRequest
    executor_class: ExecutorClass | None = None
    docker_image_name: str | None = None
    raw_script: str | None = None
    docker_run_options_preset: str
    docker_run_cmd: list[str]
    volume: Volume | None = None
    output_upload: OutputUpload | None = None

    @model_validator(mode="after")
    def validate_at_least_docker_image_or_raw_script(self) -> Self:
        if not (bool(self.docker_image_name) or bool(self.raw_script)):
            raise ValueError("Expected at least one of `docker_image_name` or `raw_script`")
        return self


class V0MachineSpecsRequest(BaseValidatorRequest, JobMixin):
    message_type: RequestType = RequestType.V0MachineSpecsRequest
    specs: MachineSpecs


class GenericError(BaseValidatorRequest):
    message_type: RequestType = RequestType.GenericError
    details: str | None = None


class ReceiptPayload(pydantic.BaseModel):
    job_uuid: str
    miner_hotkey: str
    validator_hotkey: str
    time_started: datetime.datetime
    time_took_us: int  # micro-seconds
    score_str: str

    def blob_for_signing(self):
        # pydantic v2 does not support sort_keys anymore.
        return json.dumps(self.model_dump(), sort_keys=True, default=_json_dumps_default)

    @property
    def time_took(self):
        return datetime.timedelta(microseconds=self.time_took_us)

    @property
    def score(self):
        return float(self.score_str)


class V0ReceiptRequest(BaseValidatorRequest):
    message_type: RequestType = RequestType.V0ReceiptRequest
    payload: ReceiptPayload
    signature: str

    def blob_for_signing(self):
        return self.payload.blob_for_signing()



import contextlib
import csv
import datetime
import io
import logging
import shutil
import tempfile

import bittensor
import pydantic
import requests

from .mv_protocol.validator_requests import ReceiptPayload

logger = logging.getLogger(__name__)


class Receipt(pydantic.BaseModel):
    payload: ReceiptPayload
    validator_signature: str
    miner_signature: str

    def verify_miner_signature(self):
        miner_keypair = bittensor.Keypair(ss58_address=self.payload.miner_hotkey)
        return miner_keypair.verify(self.payload.blob_for_signing(), self.miner_signature)

    def verify_validator_signature(self):
        validator_keypair = bittensor.Keypair(ss58_address=self.payload.validator_hotkey)
        return validator_keypair.verify(self.payload.blob_for_signing(), self.validator_signature)


class ReceiptFetchError(Exception):
    pass


def get_miner_receipts(hotkey: str, ip: str, port: int) -> list[Receipt]:
    """Get receipts from a given miner"""
    with contextlib.ExitStack() as exit_stack:
        try:
            receipts_url = f"http://{ip}:{port}/receipts/receipts.csv"
            response = exit_stack.enter_context(requests.get(receipts_url, stream=True, timeout=5))
            response.raise_for_status()
        except requests.RequestException as e:
            raise ReceiptFetchError("failed to get receipts from miner") from e

        temp_file = exit_stack.enter_context(tempfile.TemporaryFile())
        shutil.copyfileobj(response.raw, temp_file)
        temp_file.seek(0)

        receipts = []
        wrapper = io.TextIOWrapper(temp_file)
        csv_reader = csv.DictReader(wrapper)
        for raw_receipt in csv_reader:
            try:
                receipt = Receipt(
                    payload=ReceiptPayload(
                        job_uuid=raw_receipt["job_uuid"],
                        miner_hotkey=raw_receipt["miner_hotkey"],
                        validator_hotkey=raw_receipt["validator_hotkey"],
                        time_started=datetime.datetime.fromisoformat(raw_receipt["time_started"]),
                        time_took_us=int(raw_receipt["time_took_us"]),
                        score_str=raw_receipt["score_str"],
                    ),
                    validator_signature=raw_receipt["validator_signature"],
                    miner_signature=raw_receipt["miner_signature"],
                )
            except (ValueError, pydantic.ValidationError):
                logger.warning(f"Miner sent invalid receipt {raw_receipt=}")
                continue

            if receipt.payload.miner_hotkey != hotkey:
                logger.warning(f"Miner sent receipt of a different miner {receipt=}")
                continue

            if not receipt.verify_miner_signature():
                logger.warning(f"Invalid miner signature of receipt {receipt=}")
                continue

            if not receipt.verify_validator_signature():
                logger.warning(f"Invalid validator signature of receipt {receipt=}")
                continue

            receipts.append(receipt)

        return receipts



import datetime
from typing import TYPE_CHECKING

import bittensor
import pydantic

if TYPE_CHECKING:
    from bittensor.chain_data import NeuronInfo

BAC_VALIDATOR_SS58_ADDRESS = "5HBVrFGy6oYhhh71m9fFGYD7zbKyAeHnWN8i8s9fJTBMCtEE"
MIN_STAKE = 1000
VALIDATORS_LIMIT = 24


class MachineSpecs(pydantic.BaseModel):
    specs: dict

    def __str__(self) -> str:
        return str(self.specs)


def get_validators(netuid=12, network="finney") -> list["NeuronInfo"]:
    """
    Validators are top 24 neurons in terms of stake, only taking into account those that have at least 1000
    and forcibly including BAC_VALIDATOR_SS58_ADDRESS.
    The result is sorted.
    """
    metagraph = bittensor.metagraph(netuid=netuid, network=network)
    neurons = [
        n
        for n in metagraph.neurons
        if (n.hotkey == BAC_VALIDATOR_SS58_ADDRESS or n.stake.tao >= MIN_STAKE)
    ]
    neurons = sorted(
        neurons, key=lambda n: (n.hotkey == BAC_VALIDATOR_SS58_ADDRESS, n.stake), reverse=True
    )
    return neurons[:VALIDATORS_LIMIT]


def _json_dumps_default(obj):
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()

    raise TypeError



import enum

import pydantic

from ..base_requests import BaseRequest, JobMixin
from ..executor_class import ExecutorClass
from ..utils import MachineSpecs


class RequestType(enum.Enum):
    V0AcceptJobRequest = "V0AcceptJobRequest"
    V0DeclineJobRequest = "V0DeclineJobRequest"
    V0ExecutorManifestRequest = "V0ExecutorManifestRequest"
    V0ExecutorReadyRequest = "V0ExecutorReadyRequest"
    V0ExecutorFailedRequest = "V0ExecutorFailedRequest"
    V0JobFailedRequest = "V0JobFailedRequest"
    V0JobFinishedRequest = "V0JobFinishedRequest"
    V0MachineSpecsRequest = "V0MachineSpecsRequest"
    GenericError = "GenericError"
    UnauthorizedError = "UnauthorizedError"


class ExecutorClassManifest(pydantic.BaseModel):
    # TODO: remove support for deprecated `int` executor class
    executor_class: ExecutorClass | int
    count: int


class ExecutorManifest(pydantic.BaseModel):
    executor_classes: list[ExecutorClassManifest]

    @property
    def total_count(self) -> int:
        return sum([x.count for x in self.executor_classes])


class BaseMinerRequest(BaseRequest):
    message_type: RequestType


class V0AcceptJobRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0AcceptJobRequest


class V0DeclineJobRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0DeclineJobRequest


class V0ExecutorReadyRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0ExecutorReadyRequest


class V0ExecutorFailedRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0ExecutorFailedRequest


class V0JobFailedRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0JobFailedRequest
    docker_process_exit_status: int | None = None
    docker_process_stdout: str  # TODO: add max_length
    docker_process_stderr: str  # TODO: add max_length


class V0JobFinishedRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0JobFinishedRequest
    docker_process_stdout: str  # TODO: add max_length
    docker_process_stderr: str  # TODO: add max_length


class V0MachineSpecsRequest(BaseMinerRequest, JobMixin):
    message_type: RequestType = RequestType.V0MachineSpecsRequest
    specs: MachineSpecs


class V0ExecutorManifestRequest(BaseMinerRequest):
    message_type: RequestType = RequestType.V0ExecutorManifestRequest
    manifest: ExecutorManifest


class GenericError(BaseMinerRequest):
    message_type: RequestType = RequestType.GenericError
    details: str | None = None


class UnauthorizedErrorType(enum.Enum):
    TOKEN_TOO_OLD = "TOKEN_TOO_OLD"
    UNKNOWN_VALIDATOR = "UNKNOWN_VALIDATOR"
    VALIDATOR_INACTIVE = "VALIDATOR_INACTIVE"


class UnauthorizedError(BaseMinerRequest):
    message_type: RequestType = RequestType.UnauthorizedError
    code: UnauthorizedErrorType
    details: str | None = None



import enum

from ..base_requests import BaseRequest, JobMixin
from ..utils import MachineSpecs


class RequestType(enum.Enum):
    V0ReadyRequest = "V0ReadyRequest"
    V0FailedToPrepare = "V0FailedToPrepare"
    V0FinishedRequest = "V0FinishedRequest"
    V0FailedRequest = "V0FailedRequest"
    V0MachineSpecsRequest = "V0MachineSpecsRequest"
    GenericError = "GenericError"


class BaseExecutorRequest(BaseRequest):
    message_type: RequestType


class V0ReadyRequest(BaseExecutorRequest, JobMixin):
    message_type: RequestType = RequestType.V0ReadyRequest


class V0FailedToPrepare(BaseExecutorRequest, JobMixin):
    message_type: RequestType = RequestType.V0FailedToPrepare


class V0FailedRequest(BaseExecutorRequest, JobMixin):
    message_type: RequestType = RequestType.V0FailedRequest
    docker_process_exit_status: int | None = None
    timeout: bool
    docker_process_stdout: str  # TODO: add max_length
    docker_process_stderr: str  # TODO: add max_length


class V0MachineSpecsRequest(BaseExecutorRequest, JobMixin):
    message_type: RequestType = RequestType.V0MachineSpecsRequest
    specs: MachineSpecs


class V0FinishedRequest(BaseExecutorRequest, JobMixin):
    message_type: RequestType = RequestType.V0FinishedRequest
    docker_process_stdout: str  # TODO: add max_length
    docker_process_stderr: str  # TODO: add max_length


class GenericError(BaseExecutorRequest):
    message_type: RequestType = RequestType.GenericError
    details: str | None = None
